#pragma once

typedef unsigned char byte;
typedef unsigned long long int number;

enum SearchLimit : number
{
	// 10^15
	SQRT2 = 31622777,
	value = 1000000000000000,

	// We need to calculate all primes to factorize numbers up to S(n) where n <= SearchLimit::value
	// Since S(n) / n < 2 / 1 * 3 / 2 * 5 / 4 * ... * 47 / 46 = 7.2095926010299534105882364948723... for all 64-bit numbers, S(n) < n * 7.2095926010299534105882364948723...
	// So we need to calculate all primes up to sqrt(n * 7.2095926010299534105882364948723...) = SearchLimit::SQRT2 * 2.6850684536953528869874068825892...
	PrimesUpToSqrtLimitValue = (SQRT2 * 268507) / 100000 + 1,

	// The first prime gap larger than 256 is 436273009 - 436273291
	// So we don't need a multiplier up until then
	ShiftMultiplier = (PrimesUpToSqrtLimitValue < 436273009) ? 1 : 2,
};

template<number N> struct CompileTimePrimes;

template<> struct CompileTimePrimes<0>  { enum { value =  2 }; };
template<> struct CompileTimePrimes<1>  { enum { value =  3 }; };
template<> struct CompileTimePrimes<2>  { enum { value =  5 }; };
template<> struct CompileTimePrimes<3>  { enum { value =  7 }; };
template<> struct CompileTimePrimes<4>  { enum { value = 11 }; };
template<> struct CompileTimePrimes<5>  { enum { value = 13 }; };
template<> struct CompileTimePrimes<6>  { enum { value = 17 }; };
template<> struct CompileTimePrimes<7>  { enum { value = 19 }; };
template<> struct CompileTimePrimes<8>  { enum { value = 23 }; };
template<> struct CompileTimePrimes<9>  { enum { value = 29 }; };
template<> struct CompileTimePrimes<10> { enum { value = 31 }; };
template<> struct CompileTimePrimes<11> { enum { value = 37 }; };
template<> struct CompileTimePrimes<12> { enum { value = 41 }; };
template<> struct CompileTimePrimes<13> { enum { value = 43 }; };
template<> struct CompileTimePrimes<14> { enum { value = 47 }; };
template<> struct CompileTimePrimes<15> { enum { value = 53 }; };
template<> struct CompileTimePrimes<16> { enum { value = 59 }; };
template<> struct CompileTimePrimes<17> { enum { value = 61 }; };
template<> struct CompileTimePrimes<18> { enum { value = 67 }; };
template<> struct CompileTimePrimes<19> { enum { value = 71 }; };
template<> struct CompileTimePrimes<20> { enum { value = 73 }; };
template<> struct CompileTimePrimes<21> { enum { value = 79 }; };
template<> struct CompileTimePrimes<22> { enum { value = 83 }; };
template<> struct CompileTimePrimes<23> { enum { value = 89 }; };
template<> struct CompileTimePrimes<24> { enum { value = 97 }; };
template<> struct CompileTimePrimes<25> { enum { value = 101 }; };
template<> struct CompileTimePrimes<26> { enum { value = 103 }; };
template<> struct CompileTimePrimes<27> { enum { value = 107 }; };
template<> struct CompileTimePrimes<28> { enum { value = 109 }; };
template<> struct CompileTimePrimes<29> { enum { value = 113 }; };
template<> struct CompileTimePrimes<30> { enum { value = 127 }; };
template<> struct CompileTimePrimes<31> { enum { value = 131 }; };
template<> struct CompileTimePrimes<32> { enum { value = 137 }; };
template<> struct CompileTimePrimes<33> { enum { value = 139 }; };
template<> struct CompileTimePrimes<34> { enum { value = 149 }; };
template<> struct CompileTimePrimes<35> { enum { value = 151 }; };
template<> struct CompileTimePrimes<36> { enum { value = 157 }; };
template<> struct CompileTimePrimes<37> { enum { value = 163 }; };
template<> struct CompileTimePrimes<38> { enum { value = 167 }; };
template<> struct CompileTimePrimes<39> { enum { value = 173 }; };
template<> struct CompileTimePrimes<40> { enum { value = 179 }; };
template<> struct CompileTimePrimes<41> { enum { value = 181 }; };
template<> struct CompileTimePrimes<42> { enum { value = 191 }; };
template<> struct CompileTimePrimes<43> { enum { value = 193 }; };
template<> struct CompileTimePrimes<44> { enum { value = 197 }; };
template<> struct CompileTimePrimes<45> { enum { value = 199 }; };
template<> struct CompileTimePrimes<46> { enum { value = 211 }; };
template<> struct CompileTimePrimes<47> { enum { value = 223 }; };
template<> struct CompileTimePrimes<48> { enum { value = 227 }; };
template<> struct CompileTimePrimes<49> { enum { value = 229 }; };
template<> struct CompileTimePrimes<50> { enum { value = 233 }; };
template<> struct CompileTimePrimes<51> { enum { value = 239 }; };
template<> struct CompileTimePrimes<52> { enum { value = 241 }; };
template<> struct CompileTimePrimes<53> { enum { value = 251 }; };
template<> struct CompileTimePrimes<54> { enum { value = 257 }; };
template<> struct CompileTimePrimes<55> { enum { value = 263 }; };
template<> struct CompileTimePrimes<56> { enum { value = 269 }; };
template<> struct CompileTimePrimes<57> { enum { value = 271 }; };
template<> struct CompileTimePrimes<58> { enum { value = 277 }; };
template<> struct CompileTimePrimes<59> { enum { value = 281 }; };
template<> struct CompileTimePrimes<60> { enum { value = 283 }; };
template<> struct CompileTimePrimes<61> { enum { value = 293 }; };
template<> struct CompileTimePrimes<62> { enum { value = 307 }; };
template<> struct CompileTimePrimes<63> { enum { value = 311 }; };
template<> struct CompileTimePrimes<64> { enum { value = 313 }; };
template<> struct CompileTimePrimes<65> { enum { value = 317 }; };
template<> struct CompileTimePrimes<66> { enum { value = 331 }; };
template<> struct CompileTimePrimes<67> { enum { value = 337 }; };
template<> struct CompileTimePrimes<68> { enum { value = 347 }; };
template<> struct CompileTimePrimes<69> { enum { value = 349 }; };
template<> struct CompileTimePrimes<70> { enum { value = 353 }; };
template<> struct CompileTimePrimes<71> { enum { value = 359 }; };
template<> struct CompileTimePrimes<72> { enum { value = 367 }; };
template<> struct CompileTimePrimes<73> { enum { value = 373 }; };
template<> struct CompileTimePrimes<74> { enum { value = 379 }; };
template<> struct CompileTimePrimes<75> { enum { value = 383 }; };
template<> struct CompileTimePrimes<76> { enum { value = 389 }; };
template<> struct CompileTimePrimes<77> { enum { value = 397 }; };
template<> struct CompileTimePrimes<78> { enum { value = 401 }; };
template<> struct CompileTimePrimes<79> { enum { value = 409 }; };
template<> struct CompileTimePrimes<80> { enum { value = 419 }; };

enum { CompileTimePrimesCount = 80 };

enum PrimeTableParameters
{
	Modulo = 210,
	NumOffsets = 48,
};

static_assert(PrimeTableParameters::Modulo < 256, "Modulo is too large");

const unsigned int NumbersCoprimeToModulo[NumOffsets * 2] = {
	  1,  11,  13,  17,  19,  23,  29,  31,
	 37,  41,  43,  47,  53,  59,  61,  67,
	 71,  73,  79,  83,  89,  97, 101, 103,
	107, 109, 113, 121, 127, 131, 137, 139,
	143, 149, 151, 157, 163, 167, 169, 173,
	179, 181, 187, 191, 193, 197, 199, 209,

	  1 + 210,  11 + 210,  13 + 210,  17 + 210,  19 + 210,  23 + 210,  29 + 210,  31 + 210,
	 37 + 210,  41 + 210,  43 + 210,  47 + 210,  53 + 210,  59 + 210,  61 + 210,  67 + 210,
	 71 + 210,  73 + 210,  79 + 210,  83 + 210,  89 + 210,  97 + 210, 101 + 210, 103 + 210,
	107 + 210, 109 + 210, 113 + 210, 121 + 210, 127 + 210, 131 + 210, 137 + 210, 139 + 210,
	143 + 210, 149 + 210, 151 + 210, 157 + 210, 163 + 210, 167 + 210, 169 + 210, 173 + 210,
	179 + 210, 181 + 210, 187 + 210, 191 + 210, 193 + 210, 197 + 210, 199 + 210, 209 + 210,
};

enum Byte
{
	Bits = 8,
};

extern unsigned __int64(__fastcall *udiv128)(unsigned __int64 numhi, unsigned __int64 numlo, unsigned __int64 den, unsigned __int64* rem);

#pragma warning(push)

// Disable this warning here because number overflows are the essence of this algorithm: it computes everything modulo 2^64.
#pragma warning(disable : 4307) // warning C4307: '*' : integral constant overflow

template<number x1, number x2, number v1, number v2>
struct ExtendedEuclidCompileTime
{
	enum Variables : number
	{
		q = v1 / v2,
		x3 = x1 - q * x2,
		v3 = v1 - q * v2,
		value = ExtendedEuclidCompileTime<x2, x3, v2, v3>::value,
	};
};

template<number x1, number x2, number v1>
struct ExtendedEuclidCompileTime<x1, x2, v1, 1>
{
	enum Variables : number
	{
		value = x2,
	};
};

template<number N>
struct MultiplicativeInverse
{
	enum Variables : number
	{
		value = ExtendedEuclidCompileTime<number(-1), 1, ~N + 1, N>::value,
	};
};

#pragma warning(pop)
