#pragma once

#include "PrimeTables.h"
#include <ostream>

template<int PrimeIndex>
FORCEINLINE void FactorizeInternal(number& a, number& sumA, std::vector<std::pair<number, number>>& aFactorization)
{
	enum {p = CompileTimePrimes<PrimeIndex>::value};

	if (p * p > a)
		return;

	number q = a * MultiplicativeInverse<p>::value;
	if (q <= number(-1) / p)
	{
		std::pair<number, number> factor(p, 1);

		number n = p;
		number curSum = p + 1;
		a = q;

		while (p <= a)
		{
			q = a * MultiplicativeInverse<p>::value;
			if (q > number(-1) / p)
				break;
			++factor.second;
			n *= p;
			curSum += n;
			a = q;
		}
		sumA *= curSum;
		aFactorization.push_back(factor);
		if ((a <= SearchLimit::PrimesUpToSqrtLimitValue) && IsPrime(a))
			return;
	}

	FactorizeInternal<PrimeIndex + 1>(a, sumA, aFactorization);
}

FORCEINLINE bool DivideFast(const number aNumPrimesCheckedSoFar, const number a, const number p, number& q)
{
	if (aNumPrimesCheckedSoFar < ReciprocalsTableSize)
		return PrimeReciprocals[aNumPrimesCheckedSoFar].Divide(a, p, q);

	q = a / p;
	return ((a % p) == 0);
}

template<> FORCEINLINE void FactorizeInternal<CompileTimePrimesCount>(number& a, number& sumA, std::vector<std::pair<number, number>>& aFactorization)
{
	number numPrimesCheckedSoFar = CompileTimePrimesCount;
	const byte* shift = NextPrimeShifts + CompileTimePrimesCount - 1;
	number p = static_cast<number>(CompileTimePrimes<CompileTimePrimesCount - 1>::value) + *(shift++) * SearchLimit::ShiftMultiplier;
	while (p * p <= a)
	{
		number q;
		if (DivideFast(numPrimesCheckedSoFar, a, p, q))
		{
			std::pair<number, number> factor(p, 1);

			number n = p;
			number curSum = p + 1;
			a = q;

			while (p <= a)
			{
				if (!DivideFast(numPrimesCheckedSoFar, a, p, q))
					break;
				++factor.second;
				n *= p;
				curSum += n;
				a = q;
			}
			sumA *= curSum;
			aFactorization.push_back(factor);
			if ((a <= SearchLimit::PrimesUpToSqrtLimitValue) && IsPrime(a))
				return;
		}

		p = (p <= SearchLimit::PrimesUpToSqrtLimitValue) ? (p + *(shift++) * SearchLimit::ShiftMultiplier) : (p + 2);
		++numPrimesCheckedSoFar;
	}
}

void Factorize(number a, number& sumA, std::vector<std::pair<number, number>>& aFactorization);
void NormalizeFactorization(std::vector<std::pair<number, number>>& factorization);
void PrintFactorization(std::ostream& s, std::vector<std::pair<number, number>>& factorization);
