#pragma once

enum
{
	// There are exactly 192725 primes below 2^(64/3)
	// We can use this table for factorization when p^3 <= N < 2^64
	ReciprocalsTableSize = 192725,
};

// Reciprocals are calculated using algorithm published in http://www.agner.org/optimize/optimizing_assembly.pdf (section 16.9 "Integer division by a constant")
#pragma pack(push, 1)
struct SReciprocal
{
	number reciprocal;
	unsigned char increment;
	unsigned char shift;

	NOINLINE void Init(const number aDivisor)
	{
		unsigned long bitIndex;
		_BitScanReverse64(&bitIndex, aDivisor);
		shift = static_cast<unsigned char>(bitIndex);

		number remainder;
		number quotient = udiv128(number(1) << shift, 0, aDivisor, &remainder);
		if (remainder * 2 < aDivisor)
		{
			reciprocal = quotient;
			increment = 1;
		}
		else
		{
			reciprocal = quotient + 1;
			increment = 0;
		}
	}

	FORCEINLINE bool Divide(const number n, const number divisor, number& q) const
	{
		number highProduct;
		_umul128(n + increment, reciprocal, &highProduct);
		q = highProduct >> shift;
		return q * divisor == n;
	}

	FORCEINLINE number DivideNoRemainder(const number n) const
	{
		number highProduct;
		_umul128(n + increment, reciprocal, &highProduct);
		return highProduct >> shift;
	}
};
#pragma pack(pop)

extern SReciprocal privPrimeReciprocals[ReciprocalsTableSize];
#define PrimeReciprocals ((const SReciprocal* const)(privPrimeReciprocals))

enum
{
	ShiftTableSize = 129982215,
};

extern std::vector<byte> PrimesUpToSqrtLimit;
extern byte privNextPrimeShifts[ShiftTableSize];
#define NextPrimeShifts ((const byte* const)(privNextPrimeShifts))

FORCEINLINE number GCD(number a, number b)
{
	while (b)
	{
		const number prev_a = a;
		a = b;
		b = prev_a % b;
	}
	return a;
};

void PrimeTablesInit();
number CalculatePrimes(number aLowerBound, number anUpperBound, std::vector<byte>& anOutPrimes);

// Works only for numbers up to SearchLimit::PrimesUpToSqrtLimitValue
bool IsPrime(number n);
